# Kinesis Aggregation/Deaggregation Libraries for Python
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib

# Message aggregation protocol-specific constants
# (https://github.com/awslabs/amazon-kinesis-producer/blob/master/aggregation-format.md)
MAGIC = b'\xf3\x89\x9a\xc2'
DIGEST_SIZE = hashlib.md5().digest_size

# Kinesis Limits
# (https://docs.aws.amazon.com/kinesis/latest/APIReference/API_PutRecord.html)
MAX_BYTES_PER_RECORD = 1024*1024  # 1 MB



# Kinesis Aggregation/Deaggregation Libraries for Python
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division

import sys
import os
import inspect
import aws_kinesis_agg

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parent_dir = os.path.dirname(current_dir)
sys.path.insert(0, current_dir)
sys.path.insert(1, parent_dir)

import messages_pb2
import hashlib
import six
import threading
import copy

def _calculate_varint_size(value):
    """For an integral value represented by a varint, calculate how many bytes 
    are necessary to represent the value in a protobuf message.
    (see https://developers.google.com/protocol-buffers/docs/encoding# varints)
     
    Args:
        value (int) - The value whose varint size will be calculated
    Returns:
        The number of bytes necessary to represent the input value as a varint. (int)"""
    
    if value < 0:
        raise ValueError("Size values should not be negative.")
    
    num_bits_needed = 0
    
    if value == 0:
        num_bits_needed = 1
    else:
        # shift the value right one bit at a time until
        # there are no more '1' bits left...this counts
        # how many bits we need to represent the number
        while value > 0:
            num_bits_needed += 1
            value = value >> 1
        
    # varints only use 7 bits of the byte for the actual value
    num_varint_bytes = num_bits_needed // 7
    if num_bits_needed % 7 > 0:
        num_varint_bytes += 1
        
    return num_varint_bytes
    
    
class KeySet(object):
    """A class for tracking unique partition keys or explicit hash keys for an
    aggregated Kinesis record. Also assists in keeping track of indexes for
    their locations in the protobuf tables."""
    
    def __init__(self):
        """Create a new, empty KeySet."""
        
        self.keys = []
        self.lookup = {}

    def get_potential_index(self, key):
        """If the input key were added to this KeySet, determine what
        its resulting index would be.
        
        Args:
            key (str) - The key whose index should be calculated
        Returns:
            The integer index that this key would occupy if added to the KeySet. (int)
        """
        
        if key in self.lookup:
            return self.lookup[key]
        return len(self.keys)

    def add_key(self, key):
        """Add a new key to this KeySet.
        
        Args:
            key (str) - The key to add.
        Returns:
            A tuple of (bool,int). The bool is true if this key is not 
            already in the KeySet or false otherwise. The int indicates
            the index of the key."""
        
        if key in self.lookup:
            return False, self.lookup[key]
    
        if key not in self.lookup:
            self.lookup[key] = len(self.keys)
            
        self.keys.append(key)
        return True, len(self.keys) - 1

    def contains(self, key):
        """Check if this KeySet contains the input key.
        
        Args:
            key (str) - The key whose existence in the KeySet should be checked.
        Returns:
            True if the input key exists in this KeySet, False otherwise."""
        
        return key is not None and key in self.lookup

    def clear(self):
        """Clear all existing data from this KeySet and reset it to empty."""
        
        del self.keys[:]
        self.lookup.clear()


# Not thread-safe
class RecordAggregator(object):
    """An object to ingest Kinesis user records and optimally aggregate
    them into aggregated Kinesis records.
    
    NOTE: This object is not thread-safe."""
    
    def __init__(self, max_size=aws_kinesis_agg.MAX_BYTES_PER_RECORD):
        """Create a new empty aggregator."""

        if max_size > aws_kinesis_agg.MAX_BYTES_PER_RECORD:
            raise ValueError('Invalid max_size %d exceeds maximum value %d' %
                             (max_size, aws_kinesis_agg.MAX_BYTES_PER_RECORD))

        self.max_size = max_size
        self.current_record = AggRecord(self.max_size)
        self.callbacks = []

    def on_record_complete(self, callback, execute_on_new_thread=True):
        """A method to register a callback that will be notified (on
        a separate thread) when a fully-packed record is available.
        
        Args:
            callback - A function handle or callable object that will be called
            on a separate thread every time a new aggregated record is available
            (function or callable object).
            
            execute_on_new_thread - True if callbacks should be executed on a new
            thread, False if it should be executed on the calling thread. Defaults
            to True. (boolean)"""
        
        if callback not in self.callbacks:
            self.callbacks.append((callback, execute_on_new_thread))

    def get_num_user_records(self):
        """Returns:
            The number of user records currently aggregated in this aggregated record. (int)"""
        
        return self.current_record.get_num_user_records()

    def get_size_bytes(self):
        """Returns:
            The total number of bytes in this aggregated record (based on the size of the
            serialized record. (int)"""
        
        return self.current_record.get_size_bytes()

    def clear_record(self):
        """Clear all the user records from this aggregated record and reset it to an
        empty state."""
        
        self.current_record = AggRecord(self.max_size)

    def clear_callbacks(self):
        """Clear all the callbacks from this object that were registered with the
        on_record_complete method."""
        
        del self.callbacks[:]

    def clear_and_get(self):
        """Get the current contents of this aggregated record (whether full or not)
        as a single record and then clear the contents of this object so it can
        be re-used.  This method is useful for flushing the aggregated record when
        you need to transmit it before it is full (e.g. you're shutting down or
        haven't transmitted in a while).
        
        Returns:
            A partially-filled AggRecord or None if the aggregator is empty. (AggRecord)"""
        
        if self.get_num_user_records() == 0:
            return None
        
        out_record = self.current_record
        self.clear_record()
        return out_record

    def add_user_record(self, partition_key, data, explicit_hash_key=None):
        """Add a new user record to this aggregated record (will trigger a callback
        via onRecordComplete if aggregated record is full).
           
        Args:
            partition_key (bytes or str) - The partition key of the record to add. If pk is a
                                  string type, it will be encoded to bytes using utf-8.
            data (bytes or str) - The raw data of the record to add. If data is a string type,
                                  it will be encoded to bytes using utf-8.
            explicit_hash_key (str) - The explicit hash key of the record to add (optional)
                                      If ehk is a string type, it will be encoded to bytes
                                      using utf-8.
        Returns:
            A AggRecord if this aggregated record is full and ready to
            be transmitted or null otherwise. (AggRecord)"""

        # Attempt to add to the current aggregated record
        success = self.current_record.add_user_record(partition_key, data, explicit_hash_key)
        if success:
            # we were able to add the current data to the in-flight record
            return None
        
        # If we hit this point, aggregated record is full
        # Call all the callbacks (potentially on a separate thread)
        out_record = copy.deepcopy(self.current_record)
        for (callback, execute_on_new_thread) in self.callbacks:
            if execute_on_new_thread:
                threading.Thread(target=callback, args=(out_record,)).start()
            else:
                callback(out_record)
        
        # Current record is full so clear it out, make a new empty one and add the user record
        self.clear_record()
        self.current_record.add_user_record(partition_key, data, explicit_hash_key)
        
        return out_record

    
class AggRecord(object):
    """Represents a single aggregated Kinesis record. This Kinesis record is built
    by adding multiple user records and then serializing them to bytes using the
    Kinesis aggregated record format. This class lifts heavily from the existing 
    KPL C++ libraries found at https://github.com/awslabs/amazon-kinesis-producer.
    
    This class is NOT thread-safe.
    
    For more details on the Kinesis aggregated record format, see:
    https://github.com/awslabs/amazon-kinesis-producer/blob/master/aggregation-format.md"""
    
    def __init__(self, max_size=aws_kinesis_agg.MAX_BYTES_PER_RECORD):
        """Create a new empty aggregated record."""
        
        self.agg_record = messages_pb2.AggregatedRecord()
        self._agg_partition_key = ''
        self._agg_explicit_hash_key = ''
        self._agg_size_bytes = 0
        self.partition_keys = KeySet()
        self.explicit_hash_keys = KeySet()
        self.max_size = max_size

    def get_num_user_records(self):
        """Returns:
            The current number of user records added via the "addUserRecord(...)" method. (int)"""
        
        return len(self.agg_record.records)

    def get_size_bytes(self):
        """Returns:
            The current size in bytes of this message in its serialized form. (int)"""
        
        return len(aws_kinesis_agg.MAGIC) + self._agg_size_bytes + aws_kinesis_agg.DIGEST_SIZE

    def _serialize_to_bytes(self):
        """Serialize this record to bytes.  Has no side effects (i.e. does not affect the contents of this record object).
        
        Returns: 
            A byte array containing a aggregated Kinesis record. (binary str)"""
        
        message_body = self.agg_record.SerializeToString()

        md5_calc = hashlib.md5()
        md5_calc.update(message_body)
        calculated_digest = md5_calc.digest()
        
        return aws_kinesis_agg.MAGIC + message_body + calculated_digest

    def clear(self):
        """Clears out all records and metadata from this object so that it can be
        reused just like a fresh instance of this object."""
        
        self.agg_record = aws_kinesis_agg.kpl_pb2.AggregatedRecord()
        self._agg_partition_key = ''
        self._agg_explicit_hash_key = ''
        self._agg_size_bytes = 0
        self.partition_keys.clear()
        self.explicit_hash_keys.clear()

    def get_contents(self):
        """Get the contents of this aggregated record as members that can be used
        to call the Kinesis PutRecord or PutRecords API.  Note that this method does
        not affect the contents of this object (i.e. it has no side effects).
        
        Returns:
            A tuple of (partition key, explicit hash key, binary data) that represents
            the contents of this aggregated record. (str,str,binary str)"""
        
        agg_bytes = self._serialize_to_bytes()
        return self._agg_partition_key, self._agg_explicit_hash_key, agg_bytes

    def get_partition_key(self):
        """Get the overarching partition key for the entire aggregated record.
        
        Returns: 
            The partition key to use for the aggregated record or None if this record is empty. (str)"""
        
        if self.get_num_user_records() == 0:
            return None
        
        return self._agg_partition_key

    def get_explicit_hash_key(self):
        """Get the overarching explicit hash key for the entire aggregated record.
        
        Returns: 
            The explicit hash key to use for the aggregated record or None if this record is empty. (str)"""
        
        if self.get_num_user_records() == 0:
            return None
        
        return self._agg_explicit_hash_key

    def _calculate_record_size(self, partition_key, data, explicit_hash_key=None):
        """Based on the current size of this aggregated record, calculate what the
        new size would be if we added another user record with the specified
        parameters (used to determine when this aggregated record is full and
        can't accept any more user records).  This calculation is highly dependent
        on the Kinesis message aggregation format.
     
        Args:
            partition_key - The partition key of the new record to simulate adding (str)
            explicit_hash_key - The explicit hash key of the new record to simulate adding (str) (optional)
            data - The raw data of the new record to simulate adding (binary str)
        Returns:
            The new size of this existing record in bytes if a new user
            record with the specified parameters was added. (int)"""
        
        message_size = 0
        
        # has the partition key been added to the table of known PKs yet?
        if not self.partition_keys.contains(partition_key):
            pk_length = len(partition_key)
            message_size += 1
            message_size += _calculate_varint_size(pk_length)
            message_size += pk_length
            
        # has the explicit hash key been added to the table of known EHKs yet?
        if explicit_hash_key is not None and not self.explicit_hash_keys.contains(explicit_hash_key):
            ehk_length = len(explicit_hash_key)
            message_size += 1
            message_size += _calculate_varint_size(ehk_length)
            message_size += ehk_length
            
        # remaining calculations are for adding the new record to the list of records
            
        inner_record_size = 0
        
        # partition key field
        inner_record_size += 1
        inner_record_size += _calculate_varint_size(self.partition_keys.get_potential_index(partition_key))
        
        # explicit hash key field (this is optional)
        if explicit_hash_key is not None:
            inner_record_size += 1
            inner_record_size += _calculate_varint_size(self.explicit_hash_keys.get_potential_index(explicit_hash_key))
        
        # data field
        inner_record_size += 1
        inner_record_size += _calculate_varint_size(len(data))
        inner_record_size += len(data)
        
        message_size += 1
        message_size += _calculate_varint_size(inner_record_size)
        message_size += inner_record_size
        
        return message_size

    def add_user_record(self, partition_key, data, explicit_hash_key=None):
        """Add a new user record to this existing aggregated record if there is
        enough space (based on the defined Kinesis limits for a PutRecord call).
        
        Args:
            partition_key - The partition key of the new user record to add (bytes)
            explicit_hash_key - The explicit hash key of the new user record to add (bytes)
            data - The raw data of the new user record to add (bytes)
        Returns:
            True if the new user record was successfully added to this
            aggregated record or false if this aggregated record is too full."""

        if isinstance(partition_key, six.string_types):
            partition_key_bytes = partition_key.encode('utf-8')
        else:
            partition_key_bytes = partition_key

        if explicit_hash_key is not None and isinstance(explicit_hash_key, six.string_types):
            explicit_hash_key_bytes = explicit_hash_key.encode('utf-8')
        elif explicit_hash_key is None:
            explicit_hash_key_bytes = None
            # explicit_hash_key_bytes = AggRecord._create_explicit_hash_key(partition_key_bytes)
            # explicit_hash_key = explicit_hash_key_bytes.decode('utf-8')
        else:
            explicit_hash_key_bytes = explicit_hash_key

        if isinstance(data, six.string_types):
            data_bytes = data.encode('utf-8')
        else:
            data_bytes = data

        # Validate new record size won't overflow max size for a PutRecordRequest
        size_of_new_record = self._calculate_record_size(partition_key_bytes, data_bytes, explicit_hash_key_bytes)
        if size_of_new_record > self.max_size:
            raise ValueError('Input record (PK=%s, EHK=%s, SizeBytes=%d) too big to fit inside a single agg record.' %
                             (partition_key, explicit_hash_key, size_of_new_record))
        elif self.get_size_bytes() + size_of_new_record > self.max_size:
            return False
        
        record = self.agg_record.records.add()
        record.data = data_bytes
        
        pk_add_result = self.partition_keys.add_key(partition_key)
        if pk_add_result[0]:
            self.agg_record.partition_key_table.append(partition_key)
        record.partition_key_index = pk_add_result[1]

        if explicit_hash_key:
            ehk_add_result = self.explicit_hash_keys.add_key(explicit_hash_key)
            if ehk_add_result[0]:
                self.agg_record.explicit_hash_key_table.append(explicit_hash_key)
            record.explicit_hash_key_index = ehk_add_result[1]
        
        self._agg_size_bytes += size_of_new_record
        
        # if this is the first record, we use its partition key and hash key for the entire agg record
        if len(self.agg_record.records) == 1:
            self._agg_partition_key = partition_key
            self._agg_explicit_hash_key = explicit_hash_key
            
        return True

    @staticmethod
    def _create_explicit_hash_key(partition_key_bytes):
        """Calculate a new explicit hash key based on the input partition key
        (following the algorithm from the original KPL).
    
        Args:
            partition_key The partition key to seed the new explicit hash key with (bytes)
        Returns:
            An explicit hash key based on the input partition key generated
            using an algorithm from the original KPL."""
        
        hash_key = 0
        
        md5_calc = hashlib.md5()
        md5_calc.update(partition_key_bytes)
        pk_digest = md5_calc.digest()
        
        for i in range(0, aws_kinesis_agg.DIGEST_SIZE):
            p = int(pk_digest[i].encode('hex'), 16)
            p <<= (16 - i - 1) * 8
            hash_key += p

        return str(hash_key).encode('utf-8')



# Kinesis Aggregation/Deaggregation Libraries for Python
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import sys
import os
import inspect
import aws_kinesis_agg

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parent_dir = os.path.dirname(current_dir)
sys.path.insert(0, current_dir)
sys.path.insert(1, parent_dir)

import messages_pb2
import base64
import collections
import google.protobuf.message
import hashlib
import six
import sys


def _create_user_record(ehks, pks, mr, r, sub_seq_num):
    """Given a protobuf message record, generate a new Kinesis user record
    
    ehks - The list of explicit hash keys from the protobuf message (list of str)
    pks - The list of partition keys from the protobuf message (list of str)
    mr - A single deaggregated message record from the protobuf message (dict)
    r - The original aggregated kinesis record containing the protobuf message record (dict)
    sub_seq_num - The current subsequence number within the aggregated protobuf message (int)
    
    return value - A Kinesis user record created from a message record in the protobuf message (dict)"""
    
    explicit_hash_key = None                            
    if ehks and ehks[mr.explicit_hash_key_index] is not None:
        explicit_hash_key = ehks[mr.explicit_hash_key_index]
    partition_key = pks[mr.partition_key_index]

    new_record = {
        'kinesis': {
            'kinesisSchemaVersion': r['kinesis']['kinesisSchemaVersion'],
            'sequenceNumber': r['kinesis']['sequenceNumber'],
            'approximateArrivalTimestamp': r['kinesis']['approximateArrivalTimestamp'],
            # Fill in the new values
            'explicitHashKey': explicit_hash_key,
            'partitionKey': partition_key,
            'subSequenceNumber': sub_seq_num,
            'aggregated': True,

            # We actually re-base64-encode the data because that's how Lambda Python normally delivers records
            'data': base64.b64encode(mr.data)
        }
    }

    # Copy all the metadata from the original record (except the data-specific stuff)
    for key, value in six.iteritems(r):
        if key != 'kinesis' and key != 'data':
            new_record[key] = value

    if 'recordId' in r['kinesis']:
        new_record['kinesis']['recordId'] = r['kinesis']['recordId']

    return new_record


def _get_error_string(r, message_data, ehks, pks, ar):
    """Generate a detailed error message for when protobuf parsing fails.
    
    r - The original aggregated kinesis record containing the protobuf message record (dict)
    message_data - The raw aggregated data from the protobuf message (binary str)
    ehks - The list of explicit hash keys from the protobuf message (list)
    pks - The list of partition keys from the protobuf message (list)
    ar - The protobuf aggregated record that was being parsed when the error occurred (dict)
    
    return value - A detailed error string (str)"""
    
    error_buffer = six.StringIO()
    
    error_buffer.write('Unexpected exception during deaggregation, record was:\n')
    error_buffer.write('PKS:\n')
    for pk in pks:
        error_buffer.write('%s\n' % pk)
    error_buffer.write('EHKS:\n')
    for ehk in ehks:
        error_buffer.write('%s\n' % ehk)
    for mr in ar.records:
        error_buffer.write('Record: [hasEhk=%s, ehkIdex=%d, pkIdx=%d, dataLen=%d]\n' %
                           (ehks and ehks[mr.explicit_hash_key_index] is not None,
                            mr.explicit_hash_key_index,
                            mr.partition_key_index,
                            len(mr.data)))

    error_buffer.write('Sequence number: %s\n' % (r['kinesis']['sequenceNumber']))

    error_buffer.write('Raw data: %s\n' % (base64.b64encode(message_data)))
    
    return error_buffer.getvalue()


def _convert_from_ka_format(record):
    """Convert From Kinesis Analytics record format to Kinesis Stream record format.  Kinesis Analytics preprocessors
    use a different format for aggregated Kinesis records than Kinesis Streams.

    record - Raw Kinesis Analytics record to deaggregate. (dict)

    return value - Each yield returns a single Kinesis user record. (dict)"""

    new_record = {
        'kinesis': {
            # Kinesis Analytics doesn't pass along the kinesisSchemaVersion, so this is an educated guess
            'kinesisSchemaVersion': '1.0',
            'sequenceNumber': record['kinesisStreamRecordMetadata']['sequenceNumber'],
            'partitionKey': record['kinesisStreamRecordMetadata']['partitionKey'],
            'approximateArrivalTimestamp': record['kinesisStreamRecordMetadata']['approximateArrivalTimestamp'],
            'shardId': record['kinesisStreamRecordMetadata']['shardId'],
            'data': record['data'],
            'recordId': record['recordId']
        }
    }

    return new_record


def _convert_from_kf_format(record):
    """Convert From Kinesis Firehose record format to Kinesis Stream record format.  Kinesis Firehose transformers
    use a different format for aggregated Kinesis records than Kinesis Streams.

    record - Raw Kinesis Firehose record to deaggregate. (dict)

    return value - Each yield returns a single Kinesis user record. (dict)"""

    new_record = {
        'kinesis': {
            # Kinesis Firehose doesn't pass along the kinesisSchemaVersion, so this is an educated guess
            'kinesisSchemaVersion': '1.0',
            'sequenceNumber': record['kinesisRecordMetadata']['sequenceNumber'],
            'partitionKey': record['kinesisRecordMetadata']['partitionKey'],
            'approximateArrivalTimestamp': record['kinesisRecordMetadata']['approximateArrivalTimestamp'],
            'shardId': record['kinesisRecordMetadata']['shardId'],
            'data': record['data'],
            'recordId': record['recordId']
        }
    }

    return new_record


def _convert_from_boto3_format(record):
    """Convert From Boto3 Kinesis client record format to Kinesis Stream record format.

    record - Raw Boto3 Kinesis client record to deaggregate. (dict)

    return value - Each yield returns a single Kinesis user record. (dict)"""

    new_record = {
        'kinesis': {
            'kinesisSchemaVersion': '1.0',
            'sequenceNumber': record['SequenceNumber'],
            'partitionKey': record['PartitionKey'],
            'approximateArrivalTimestamp': record['ApproximateArrivalTimestamp'],
            'data': record['Data']
        }
    }

    return new_record


def deaggregate_records(records):
    """Given a set of Kinesis records, deaggregate any records that were packed using the
    Kinesis Producer Library into individual records.  This method will be a no-op for any
    records that are not aggregated (but will still return them).
    
    records - The list of raw Kinesis records to deaggregate. (list of dict)
    
    return value - A list of Kinesis user records greater than or equal to the size of the 
    input record list. (list of dict)"""
    
    # Use the existing generator function to deaggregate all the records
    return_records = []
    return_records.extend(iter_deaggregate_records(records))        
    return return_records


def iter_deaggregate_records(records, data_format=None):
    """Generator function - Given a set of Kinesis records, deaggregate them one at a time
    using the Kinesis aggregated message format.  This method will not affect any
    records that are not aggregated (but will still return them).
    
    records - The list of raw Kinesis records to deaggregate. (list of dict)
    
    return value - Each yield returns a single Kinesis user record. (dict)"""
    
    # We got a single record...try to coerce it to a list
    if isinstance(records, collections.abc.Mapping):
        records = [records]
        
    for r in records:
        is_aggregated = True
        sub_seq_num = 0

        if 'kinesis' not in r and 'data' in r or 'Data' in r:
            # Kinesis Analytics preprocessors & Firehose transformers use a different format for aggregated
            # Kinesis Stream records, so we're going to convert KA / KF style records to KS style records.
            if 'kinesisStreamRecordMetadata' in r:
                # Kinesis Analytics style record
                r = _convert_from_ka_format(r)
            elif 'kinesisRecordMetadata' in r:
                # Kinesis Firehose style record
                r = _convert_from_kf_format(r)
            elif data_format == 'Boto3':
                # Boto3 Kinesis client style record
                r = _convert_from_boto3_format(r)

        # Decode the incoming data
        raw_data = r['kinesis']['data']

        decoded_data = base64.b64decode(raw_data) if data_format != 'Boto3' else raw_data
        
        # Verify the magic header
        data_magic = None
        if len(decoded_data) >= len(aws_kinesis_agg.MAGIC):
            data_magic = decoded_data[:len(aws_kinesis_agg.MAGIC)]
        else:
            is_aggregated = False
        
        decoded_data_no_magic = decoded_data[len(aws_kinesis_agg.MAGIC):]
        
        if aws_kinesis_agg.MAGIC != data_magic or len(decoded_data_no_magic) <= aws_kinesis_agg.DIGEST_SIZE:
            is_aggregated = False
            
        if is_aggregated:            
            
            # verify the MD5 digest
            message_digest = decoded_data_no_magic[-aws_kinesis_agg.DIGEST_SIZE:]
            message_data = decoded_data_no_magic[:-aws_kinesis_agg.DIGEST_SIZE]

            md5_calc = hashlib.md5()
            md5_calc.update(message_data)
            calculated_digest = md5_calc.digest()
            
            if message_digest != calculated_digest:            
                is_aggregated = False            
            else:                            
                # Extract the protobuf message
                try:    
                    ar = messages_pb2.AggregatedRecord()
                    ar.ParseFromString(message_data)
                    
                    pks = ar.partition_key_table
                    ehks = ar.explicit_hash_key_table
                    
                    try:                    
                        # Split out all the aggregated records into individual records    
                        for mr in ar.records:                                                    
                            new_record = _create_user_record(ehks, pks, mr, r, sub_seq_num)
                            sub_seq_num += 1
                            yield new_record
                        
                    except Exception as e:        
                        
                        error_string = _get_error_string(r, message_data, ehks, pks, ar)
                        print('ERROR: %s\n%s' % (str(e), error_string), file=sys.stderr)
                    
                except google.protobuf.message.DecodeError:                    
                    is_aggregated = False
        
        if not is_aggregated:
            yield r
    
    return




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: messages.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import config_pb2 as config__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0emessages.proto\x12\x14\x61ws.kinesis.protobuf\x1a\x0c\x63onfig.proto\"!\n\x03Tag\x12\x0b\n\x03key\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x01(\t\"}\n\x06Record\x12\x1b\n\x13partition_key_index\x18\x01 \x02(\x04\x12\x1f\n\x17\x65xplicit_hash_key_index\x18\x02 \x01(\x04\x12\x0c\n\x04\x64\x61ta\x18\x03 \x02(\x0c\x12\'\n\x04tags\x18\x04 \x03(\x0b\x32\x19.aws.kinesis.protobuf.Tag\"\x7f\n\x10\x41ggregatedRecord\x12\x1b\n\x13partition_key_table\x18\x01 \x03(\t\x12\x1f\n\x17\x65xplicit_hash_key_table\x18\x02 \x03(\t\x12-\n\x07records\x18\x03 \x03(\x0b\x32\x1c.aws.kinesis.protobuf.Record\"\xe6\x03\n\x07Message\x12\n\n\x02id\x18\x01 \x02(\x04\x12\x11\n\tsource_id\x18\x02 \x01(\x04\x12\x35\n\nput_record\x18\x03 \x01(\x0b\x32\x1f.aws.kinesis.protobuf.PutRecordH\x00\x12,\n\x05\x66lush\x18\x04 \x01(\x0b\x32\x1b.aws.kinesis.protobuf.FlushH\x00\x12\x42\n\x11put_record_result\x18\x05 \x01(\x0b\x32%.aws.kinesis.protobuf.PutRecordResultH\x00\x12<\n\rconfiguration\x18\x06 \x01(\x0b\x32#.aws.kinesis.protobuf.ConfigurationH\x00\x12?\n\x0fmetrics_request\x18\x07 \x01(\x0b\x32$.aws.kinesis.protobuf.MetricsRequestH\x00\x12\x41\n\x10metrics_response\x18\x08 \x01(\x0b\x32%.aws.kinesis.protobuf.MetricsResponseH\x00\x12?\n\x0fset_credentials\x18\t \x01(\x0b\x32$.aws.kinesis.protobuf.SetCredentialsH\x00\x42\x10\n\x0e\x61\x63tual_message\"`\n\tPutRecord\x12\x13\n\x0bstream_name\x18\x01 \x02(\t\x12\x15\n\rpartition_key\x18\x02 \x02(\t\x12\x19\n\x11\x65xplicit_hash_key\x18\x03 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x04 \x02(\x0c\"\x1c\n\x05\x46lush\x12\x13\n\x0bstream_name\x18\x01 \x01(\t\"f\n\x07\x41ttempt\x12\r\n\x05\x64\x65lay\x18\x01 \x02(\r\x12\x10\n\x08\x64uration\x18\x02 \x02(\r\x12\x0f\n\x07success\x18\x03 \x02(\x08\x12\x12\n\nerror_code\x18\x04 \x01(\t\x12\x15\n\rerror_message\x18\x05 \x01(\t\"~\n\x0fPutRecordResult\x12/\n\x08\x61ttempts\x18\x01 \x03(\x0b\x32\x1d.aws.kinesis.protobuf.Attempt\x12\x0f\n\x07success\x18\x02 \x02(\x08\x12\x10\n\x08shard_id\x18\x03 \x01(\t\x12\x17\n\x0fsequence_number\x18\x04 \x01(\t\">\n\x0b\x43redentials\x12\x0c\n\x04\x61kid\x18\x01 \x02(\t\x12\x12\n\nsecret_key\x18\x02 \x02(\t\x12\r\n\x05token\x18\x03 \x01(\t\"]\n\x0eSetCredentials\x12\x13\n\x0b\x66or_metrics\x18\x01 \x01(\x08\x12\x36\n\x0b\x63redentials\x18\x02 \x02(\x0b\x32!.aws.kinesis.protobuf.Credentials\"\'\n\tDimension\x12\x0b\n\x03key\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x02(\t\"K\n\x05Stats\x12\r\n\x05\x63ount\x18\x01 \x02(\x01\x12\x0b\n\x03sum\x18\x02 \x02(\x01\x12\x0c\n\x04mean\x18\x03 \x02(\x01\x12\x0b\n\x03min\x18\x04 \x02(\x01\x12\x0b\n\x03max\x18\x05 \x02(\x01\"\x88\x01\n\x06Metric\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x33\n\ndimensions\x18\x02 \x03(\x0b\x32\x1f.aws.kinesis.protobuf.Dimension\x12*\n\x05stats\x18\x03 \x02(\x0b\x32\x1b.aws.kinesis.protobuf.Stats\x12\x0f\n\x07seconds\x18\x04 \x02(\x04\"/\n\x0eMetricsRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07seconds\x18\x02 \x01(\x04\"@\n\x0fMetricsResponse\x12-\n\x07metrics\x18\x01 \x03(\x0b\x32\x1c.aws.kinesis.protobuf.MetricB2\n0com.amazonaws.services.kinesis.producer.protobuf')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'messages_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n0com.amazonaws.services.kinesis.producer.protobuf'
  _TAG._serialized_start=54
  _TAG._serialized_end=87
  _RECORD._serialized_start=89
  _RECORD._serialized_end=214
  _AGGREGATEDRECORD._serialized_start=216
  _AGGREGATEDRECORD._serialized_end=343
  _MESSAGE._serialized_start=346
  _MESSAGE._serialized_end=832
  _PUTRECORD._serialized_start=834
  _PUTRECORD._serialized_end=930
  _FLUSH._serialized_start=932
  _FLUSH._serialized_end=960
  _ATTEMPT._serialized_start=962
  _ATTEMPT._serialized_end=1064
  _PUTRECORDRESULT._serialized_start=1066
  _PUTRECORDRESULT._serialized_end=1192
  _CREDENTIALS._serialized_start=1194
  _CREDENTIALS._serialized_end=1256
  _SETCREDENTIALS._serialized_start=1258
  _SETCREDENTIALS._serialized_end=1351
  _DIMENSION._serialized_start=1353
  _DIMENSION._serialized_end=1392
  _STATS._serialized_start=1394
  _STATS._serialized_end=1469
  _METRIC._serialized_start=1472
  _METRIC._serialized_end=1608
  _METRICSREQUEST._serialized_start=1610
  _METRICSREQUEST._serialized_end=1657
  _METRICSRESPONSE._serialized_start=1659
  _METRICSRESPONSE._serialized_end=1723
# @@protoc_insertion_point(module_scope)



# Kinesis Aggregation/Deaggregation Libraries for Python
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import open
from setuptools import setup

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='aws_kinesis_agg',
    packages=['aws_kinesis_agg'],
    version='1.2.2',
    description='Python module to assist in taking advantage of the Kinesis message aggregation '
                'format for both aggregation and deaggregation.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Ian Meyers',
    author_email='meyersi@amazon.com',
    license="Apache-2.0",
    url='http://github.com/awslabs/kinesis-aggregation',
    keywords=['aws', 'kinesis', 'aggregation', 'deaggregation', 'kpl'],
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: Apache Software License',
                 'Natural Language :: English',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.6',
                 'Topic :: Software Development :: Libraries :: Python Modules',
                 'Topic :: Utilities'],
    install_requires=['protobuf']
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: config.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0c\x63onfig.proto\x12\x14\x61ws.kinesis.protobuf\"F\n\x13\x41\x64\x64itionalDimension\x12\x0b\n\x03key\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x02(\t\x12\x13\n\x0bgranularity\x18\x03 \x02(\t\"\x90\t\n\rConfiguration\x12J\n\x16\x61\x64\x64itional_metric_dims\x18\x80\x01 \x03(\x0b\x32).aws.kinesis.protobuf.AdditionalDimension\x12!\n\x13\x61ggregation_enabled\x18\x01 \x01(\x08:\x04true\x12)\n\x15\x61ggregation_max_count\x18\x02 \x01(\x04:\n4294967295\x12#\n\x14\x61ggregation_max_size\x18\x03 \x01(\x04:\x05\x35\x31\x32\x30\x30\x12\x1d\n\x13\x63loudwatch_endpoint\x18\x04 \x01(\t:\x00\x12\x1c\n\x0f\x63loudwatch_port\x18\x05 \x01(\x04:\x03\x34\x34\x33\x12!\n\x14\x63ollection_max_count\x18\x06 \x01(\x04:\x03\x35\x30\x30\x12$\n\x13\x63ollection_max_size\x18\x07 \x01(\x04:\x07\x35\x32\x34\x32\x38\x38\x30\x12\x1d\n\x0f\x63onnect_timeout\x18\x08 \x01(\x04:\x04\x36\x30\x30\x30\x12 \n\x11\x65nable_core_dumps\x18\t \x01(\x08:\x05\x66\x61lse\x12 \n\x11\x66\x61il_if_throttled\x18\n \x01(\x08:\x05\x66\x61lse\x12\x1a\n\x10kinesis_endpoint\x18\x0b \x01(\t:\x00\x12\x19\n\x0ckinesis_port\x18\x0c \x01(\x04:\x03\x34\x34\x33\x12\x17\n\tlog_level\x18\r \x01(\t:\x04info\x12\x1b\n\x0fmax_connections\x18\x0e \x01(\x04:\x02\x32\x34\x12\"\n\x13metrics_granularity\x18\x0f \x01(\t:\x05shard\x12\x1f\n\rmetrics_level\x18\x10 \x01(\t:\x08\x64\x65tailed\x12\x31\n\x11metrics_namespace\x18\x11 \x01(\t:\x16KinesisProducerLibrary\x12#\n\x14metrics_upload_delay\x18\x12 \x01(\x04:\x05\x36\x30\x30\x30\x30\x12\x1a\n\x0fmin_connections\x18\x13 \x01(\x04:\x01\x31\x12\x17\n\nrate_limit\x18\x14 \x01(\x04:\x03\x31\x35\x30\x12%\n\x18record_max_buffered_time\x18\x15 \x01(\x04:\x03\x31\x30\x30\x12\x19\n\nrecord_ttl\x18\x16 \x01(\x04:\x05\x33\x30\x30\x30\x30\x12\x10\n\x06region\x18\x17 \x01(\t:\x00\x12\x1d\n\x0frequest_timeout\x18\x18 \x01(\x04:\x04\x36\x30\x30\x30\x12 \n\x12verify_certificate\x18\x19 \x01(\x08:\x04true\x12\x14\n\nproxy_host\x18\x1a \x01(\t:\x00\x12\x17\n\nproxy_port\x18\x1b \x01(\x04:\x03\x34\x34\x33\x12\x19\n\x0fproxy_user_name\x18\x1c \x01(\t:\x00\x12\x18\n\x0eproxy_password\x18\x1d \x01(\t:\x00\x12T\n\rthread_config\x18\x1e \x01(\x0e\x32\x30.aws.kinesis.protobuf.Configuration.ThreadConfig:\x0bPER_REQUEST\x12\x1c\n\x10thread_pool_size\x18\x1f \x01(\r:\x02\x36\x34\"+\n\x0cThreadConfig\x12\x0f\n\x0bPER_REQUEST\x10\x00\x12\n\n\x06POOLED\x10\x01\x42\x32\n0com.amazonaws.services.kinesis.producer.protobuf')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'config_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n0com.amazonaws.services.kinesis.producer.protobuf'
  _ADDITIONALDIMENSION._serialized_start=38
  _ADDITIONALDIMENSION._serialized_end=108
  _CONFIGURATION._serialized_start=111
  _CONFIGURATION._serialized_end=1279
  _CONFIGURATION_THREADCONFIG._serialized_start=1236
  _CONFIGURATION_THREADCONFIG._serialized_end=1279
# @@protoc_insertion_point(module_scope)
